@@ -31,7 +31,7 @@ def infer_decorator_signature_if_simple(dec: Decorator,
                 [None],
                 AnyType(TypeOfAny.special_form),
                 analyzer.named_type('__builtins__.function'),
-                name=dec.var.name())
+                name=dec.var.name)
         elif isinstance(dec.func.type, CallableType):
             dec.var.type = dec.func.type
         return
