@@ -4,29 +4,35 @@ The top-level AST transformation logic is implemented in mypyc.irbuild.visitor
 and mypyc.irbuild.builder.
 """
 
-from typing import List, Optional, Union
+from typing import List, Optional, Union, Callable, cast
 
 from mypy.nodes import (
     Expression, NameExpr, MemberExpr, SuperExpr, CallExpr, UnaryExpr, OpExpr, IndexExpr,
     ConditionalExpr, ComparisonExpr, IntExpr, FloatExpr, ComplexExpr, StrExpr,
     BytesExpr, EllipsisExpr, ListExpr, TupleExpr, DictExpr, SetExpr, ListComprehension,
     SetComprehension, DictionaryComprehension, SliceExpr, GeneratorExpr, CastExpr, StarExpr,
+    AssignmentExpr,
     Var, RefExpr, MypyFile, TypeInfo, TypeApplication, LDEF, ARG_POS
 )
-from mypy.types import TupleType, get_proper_type
+from mypy.types import TupleType, get_proper_type, Instance
 
+from mypyc.common import MAX_SHORT_INT
 from mypyc.ir.ops import (
-    Value, TupleGet, TupleSet, PrimitiveOp, BasicBlock, OpDescription, Assign
+    Value, Register, TupleGet, TupleSet, BasicBlock, Assign, LoadAddress
+)
+from mypyc.ir.rtypes import (
+    RTuple, object_rprimitive, is_none_rprimitive, int_rprimitive, is_int_rprimitive
 )
-from mypyc.ir.rtypes import RTuple, object_rprimitive, is_none_rprimitive
 from mypyc.ir.func_ir import FUNC_CLASSMETHOD, FUNC_STATICMETHOD
-from mypyc.primitives.registry import name_ref_ops
+from mypyc.primitives.registry import CFunctionDescription, builtin_names
 from mypyc.primitives.generic_ops import iter_op
-from mypyc.primitives.misc_ops import new_slice_op, ellipsis_op, type_op
-from mypyc.primitives.list_ops import new_list_op, list_append_op, list_extend_op
-from mypyc.primitives.tuple_ops import list_tuple_op
-from mypyc.primitives.dict_ops import new_dict_op, dict_set_item_op
+from mypyc.primitives.misc_ops import new_slice_op, ellipsis_op, type_op, get_module_dict_op
+from mypyc.primitives.list_ops import list_append_op, list_extend_op, list_slice_op
+from mypyc.primitives.tuple_ops import list_tuple_op, tuple_slice_op
+from mypyc.primitives.dict_ops import dict_new_op, dict_set_item_op, dict_get_item_op
 from mypyc.primitives.set_ops import new_set_op, set_add_op, set_update_op
+from mypyc.primitives.str_ops import str_slice_op
+from mypyc.primitives.int_ops import int_comparison_op_mapping
 from mypyc.irbuild.specialize import specializers
 from mypyc.irbuild.builder import IRBuilder
 from mypyc.irbuild.for_helpers import translate_list_comprehension, comprehension_helper
@@ -38,11 +44,16 @@ from mypyc.irbuild.for_helpers import translate_list_comprehension, comprehensio
 def transform_name_expr(builder: IRBuilder, expr: NameExpr) -> Value:
     assert expr.node, "RefExpr not resolved"
     fullname = expr.node.fullname
-    if fullname in name_ref_ops:
-        # Use special access op for this particular name.
-        desc = name_ref_ops[fullname]
-        assert desc.result_type is not None
-        return builder.add(PrimitiveOp([], desc, expr.line))
+    if fullname in builtin_names:
+        typ, src = builtin_names[fullname]
+        return builder.add(LoadAddress(typ, src, expr.line))
+    # special cases
+    if fullname == 'builtins.None':
+        return builder.none()
+    if fullname == 'builtins.True':
+        return builder.true()
+    if fullname == 'builtins.False':
+        return builder.false()
 
     if isinstance(expr.node, Var) and expr.node.is_final:
         value = builder.emit_load_final(
@@ -74,8 +85,21 @@ def transform_name_expr(builder: IRBuilder, expr: NameExpr) -> Value:
                     expr.node.name),
                 expr.node.line)
 
-        # TODO: Behavior currently only defined for Var and FuncDef node types.
-        return builder.read(builder.get_assignment_target(expr), expr.line)
+        # TODO: Behavior currently only defined for Var, FuncDef and MypyFile node types.
+        if isinstance(expr.node, MypyFile):
+            # Load reference to a module imported inside function from
+            # the modules dictionary. It would be closer to Python
+            # semantics to access modules imported inside functions
+            # via local variables, but this is tricky since the mypy
+            # AST doesn't include a Var node for the module. We
+            # instead load the module separately on each access.
+            mod_dict = builder.call_c(get_module_dict_op, [], expr.line)
+            obj = builder.call_c(dict_get_item_op,
+                                 [mod_dict, builder.load_static_unicode(expr.node.fullname)],
+                                 expr.line)
+            return obj
+        else:
+            return builder.read(builder.get_assignment_target(expr), expr.line)
 
     return builder.load_global(expr)
 
@@ -114,11 +138,12 @@ def transform_super_expr(builder: IRBuilder, o: SuperExpr) -> Value:
         assert o.info is not None
         typ = builder.load_native_type_object(o.info.fullname)
         ir = builder.mapper.type_to_ir[o.info]
-        iter_env = iter(builder.environment.indexes)
-        vself = next(iter_env)  # grab first argument
+        iter_env = iter(builder.builder.args)
+        # Grab first argument
+        vself = next(iter_env)  # type: Value
         if builder.fn_info.is_generator:
             # grab sixth argument (see comment in translate_super_method_call)
-            self_targ = list(builder.environment.symtable.values())[6]
+            self_targ = list(builder.symtables[-1].values())[6]
             vself = builder.read(self_targ, builder.fn_info.fitem.line)
         elif not ir.is_ext_class:
             vself = next(iter_env)  # second argument is self if non_extension class
@@ -241,8 +266,28 @@ def translate_method_call(builder: IRBuilder, expr: CallExpr, callee: MemberExpr
 
 
 def translate_super_method_call(builder: IRBuilder, expr: CallExpr, callee: SuperExpr) -> Value:
-    if callee.info is None or callee.call.args:
+    if callee.info is None or (len(callee.call.args) != 0 and len(callee.call.args) != 2):
         return translate_call(builder, expr, callee)
+
+    # We support two-argument super but only when it is super(CurrentClass, self)
+    # TODO: We could support it when it is a parent class in many cases?
+    if len(callee.call.args) == 2:
+        self_arg = callee.call.args[1]
+        if (
+            not isinstance(self_arg, NameExpr)
+            or not isinstance(self_arg.node, Var)
+            or not self_arg.node.is_self
+        ):
+            return translate_call(builder, expr, callee)
+
+        typ_arg = callee.call.args[0]
+        if (
+            not isinstance(typ_arg, NameExpr)
+            or not isinstance(typ_arg.node, TypeInfo)
+            or callee.info is not typ_arg.node
+        ):
+            return translate_call(builder, expr, callee)
+
     ir = builder.mapper.type_to_ir[callee.info]
     # Search for the method in the mro, skipping ourselves.
     for base in ir.mro[1:]:
@@ -256,16 +301,17 @@ def translate_super_method_call(builder: IRBuilder, expr: CallExpr, callee: Supe
     arg_kinds, arg_names = expr.arg_kinds[:], expr.arg_names[:]
 
     if decl.kind != FUNC_STATICMETHOD:
-        vself = next(iter(builder.environment.indexes))  # grab first argument
+        # Grab first argument
+        vself = builder.self()  # type: Value
         if decl.kind == FUNC_CLASSMETHOD:
-            vself = builder.primitive_op(type_op, [vself], expr.line)
+            vself = builder.call_c(type_op, [vself], expr.line)
         elif builder.fn_info.is_generator:
             # For generator classes, the self target is the 6th value
             # in the symbol table (which is an ordered dict). This is sort
             # of ugly, but we can't search by name since the 'self' parameter
             # could be named anything, and it doesn't get added to the
             # environment indexes.
-            self_targ = list(builder.environment.symtable.values())[6]
+            self_targ = list(builder.symtables[-1].values())[6]
             vself = builder.read(self_targ, builder.fn_info.fitem.line)
         arg_values.insert(0, vself)
         arg_kinds.insert(0, ARG_POS)
@@ -297,22 +343,66 @@ def transform_op_expr(builder: IRBuilder, expr: OpExpr) -> Value:
 
 def transform_index_expr(builder: IRBuilder, expr: IndexExpr) -> Value:
     base = builder.accept(expr.base)
+    index = expr.index
 
-    if isinstance(base.type, RTuple) and isinstance(expr.index, IntExpr):
-        return builder.add(TupleGet(base, expr.index.value, expr.line))
+    if isinstance(base.type, RTuple) and isinstance(index, IntExpr):
+        return builder.add(TupleGet(base, index.value, expr.line))
+
+    if isinstance(index, SliceExpr):
+        value = try_gen_slice_op(builder, base, index)
+        if value:
+            return value
 
     index_reg = builder.accept(expr.index)
     return builder.gen_method_call(
         base, '__getitem__', [index_reg], builder.node_type(expr), expr.line)
 
 
+def try_gen_slice_op(builder: IRBuilder, base: Value, index: SliceExpr) -> Optional[Value]:
+    """Generate specialized slice op for some index expressions.
+
+    Return None if a specialized op isn't available.
+
+    This supports obj[x:y], obj[:x], and obj[x:] for a few types.
+    """
+    if index.stride:
+        # We can only handle the default stride of 1.
+        return None
+
+    if index.begin_index:
+        begin_type = builder.node_type(index.begin_index)
+    else:
+        begin_type = int_rprimitive
+    if index.end_index:
+        end_type = builder.node_type(index.end_index)
+    else:
+        end_type = int_rprimitive
+
+    # Both begin and end index must be int (or missing).
+    if is_int_rprimitive(begin_type) and is_int_rprimitive(end_type):
+        if index.begin_index:
+            begin = builder.accept(index.begin_index)
+        else:
+            begin = builder.load_static_int(0)
+        if index.end_index:
+            end = builder.accept(index.end_index)
+        else:
+            # Replace missing end index with the largest short integer
+            # (a sequence can't be longer).
+            end = builder.load_static_int(MAX_SHORT_INT)
+        candidates = [list_slice_op, tuple_slice_op, str_slice_op]
+        return builder.builder.matching_call_c(candidates, [base, begin, end], index.line)
+
+    return None
+
+
 def transform_conditional_expr(builder: IRBuilder, expr: ConditionalExpr) -> Value:
     if_body, else_body, next = BasicBlock(), BasicBlock(), BasicBlock()
 
     builder.process_conditional(expr.cond, if_body, else_body)
     expr_type = builder.node_type(expr)
     # Having actual Phi nodes would be really nice here!
-    target = builder.alloc_temp(expr_type)
+    target = Register(expr_type)
 
     builder.activate_block(if_body)
     true_value = builder.accept(expr.if_expr)
@@ -332,8 +422,56 @@ def transform_conditional_expr(builder: IRBuilder, expr: ConditionalExpr) -> Val
 
 
 def transform_comparison_expr(builder: IRBuilder, e: ComparisonExpr) -> Value:
-    # TODO: Don't produce an expression when used in conditional context
+    # x in (...)/[...]
+    # x not in (...)/[...]
+    if (e.operators[0] in ['in', 'not in']
+            and len(e.operators) == 1
+            and isinstance(e.operands[1], (TupleExpr, ListExpr))):
+        items = e.operands[1].items
+        n_items = len(items)
+        # x in y -> x == y[0] or ... or x == y[n]
+        # x not in y -> x != y[0] and ... and x != y[n]
+        # 16 is arbitrarily chosen to limit code size
+        if 1 < n_items < 16:
+            if e.operators[0] == 'in':
+                bin_op = 'or'
+                cmp_op = '=='
+            else:
+                bin_op = 'and'
+                cmp_op = '!='
+            lhs = e.operands[0]
+            mypy_file = builder.graph['builtins'].tree
+            assert mypy_file is not None
+            bool_type = Instance(cast(TypeInfo, mypy_file.names['bool'].node), [])
+            exprs = []
+            for item in items:
+                expr = ComparisonExpr([cmp_op], [lhs, item])
+                builder.types[expr] = bool_type
+                exprs.append(expr)
 
+            or_expr = exprs.pop(0)  # type: Expression
+            for expr in exprs:
+                or_expr = OpExpr(bin_op, or_expr, expr)
+                builder.types[or_expr] = bool_type
+            return builder.accept(or_expr)
+        # x in [y]/(y) -> x == y
+        # x not in [y]/(y) -> x != y
+        elif n_items == 1:
+            if e.operators[0] == 'in':
+                cmp_op = '=='
+            else:
+                cmp_op = '!='
+            e.operators = [cmp_op]
+            e.operands[1] = items[0]
+        # x in []/() -> False
+        # x not in []/() -> True
+        elif n_items == 0:
+            if e.operators[0] == 'in':
+                return builder.false()
+            else:
+                return builder.true()
+
+    # TODO: Don't produce an expression when used in conditional context
     # All of the trickiness here is due to support for chained conditionals
     # (`e1 < e2 > e3`, etc). `e1 < e2 > e3` is approximately equivalent to
     # `e1 < e2 and e2 > e3` except that `e2` is only evaluated once.
@@ -362,6 +500,9 @@ def transform_basic_comparison(builder: IRBuilder,
                                left: Value,
                                right: Value,
                                line: int) -> Value:
+    if (is_int_rprimitive(left.type) and is_int_rprimitive(right.type)
+            and op in int_comparison_op_mapping.keys()):
+        return builder.compare_tagged(left, right, op, line)
     negate = False
     if op == 'is not':
         op, negate = 'is', True
@@ -400,7 +541,7 @@ def transform_bytes_expr(builder: IRBuilder, expr: BytesExpr) -> Value:
 
 
 def transform_ellipsis(builder: IRBuilder, o: EllipsisExpr) -> Value:
-    return builder.primitive_op(ellipsis_op, [], o.line)
+    return builder.add(LoadAddress(ellipsis_op.type, ellipsis_op.src, o.line))
 
 
 # Display expressions
@@ -414,10 +555,11 @@ def _visit_list_display(builder: IRBuilder, items: List[Expression], line: int)
     return _visit_display(
         builder,
         items,
-        new_list_op,
+        builder.new_list_op,
         list_append_op,
         list_extend_op,
-        line
+        line,
+        True
     )
 
 
@@ -443,7 +585,7 @@ def transform_tuple_expr(builder: IRBuilder, expr: TupleExpr) -> Value:
 def _visit_tuple_display(builder: IRBuilder, expr: TupleExpr) -> Value:
     """Create a list, then turn it into a tuple."""
     val_as_list = _visit_list_display(builder, expr.items, expr.line)
-    return builder.primitive_op(list_tuple_op, [val_as_list], expr.line)
+    return builder.call_c(list_tuple_op, [val_as_list], expr.line)
 
 
 def transform_dict_expr(builder: IRBuilder, expr: DictExpr) -> Value:
@@ -461,19 +603,21 @@ def transform_set_expr(builder: IRBuilder, expr: SetExpr) -> Value:
     return _visit_display(
         builder,
         expr.items,
-        new_set_op,
+        builder.new_set_op,
         set_add_op,
         set_update_op,
-        expr.line
+        expr.line,
+        False
     )
 
 
 def _visit_display(builder: IRBuilder,
                    items: List[Expression],
-                   constructor_op: OpDescription,
-                   append_op: OpDescription,
-                   extend_op: OpDescription,
-                   line: int
+                   constructor_op: Callable[[List[Value], int], Value],
+                   append_op: CFunctionDescription,
+                   extend_op: CFunctionDescription,
+                   line: int,
+                   is_list: bool
                    ) -> Value:
     accepted_items = []
     for item in items:
@@ -485,17 +629,17 @@ def _visit_display(builder: IRBuilder,
     result = None  # type: Union[Value, None]
     initial_items = []
     for starred, value in accepted_items:
-        if result is None and not starred and constructor_op.is_var_arg:
+        if result is None and not starred and is_list:
             initial_items.append(value)
             continue
 
         if result is None:
-            result = builder.primitive_op(constructor_op, initial_items, line)
+            result = constructor_op(initial_items, line)
 
-        builder.primitive_op(extend_op if starred else append_op, [result, value], line)
+        builder.call_c(extend_op if starred else append_op, [result, value], line)
 
     if result is None:
-        result = builder.primitive_op(constructor_op, initial_items, line)
+        result = constructor_op(initial_items, line)
 
     return result
 
@@ -509,25 +653,25 @@ def transform_list_comprehension(builder: IRBuilder, o: ListComprehension) -> Va
 
 def transform_set_comprehension(builder: IRBuilder, o: SetComprehension) -> Value:
     gen = o.generator
-    set_ops = builder.primitive_op(new_set_op, [], o.line)
+    set_ops = builder.call_c(new_set_op, [], o.line)
     loop_params = list(zip(gen.indices, gen.sequences, gen.condlists))
 
     def gen_inner_stmts() -> None:
         e = builder.accept(gen.left_expr)
-        builder.primitive_op(set_add_op, [set_ops, e], o.line)
+        builder.call_c(set_add_op, [set_ops, e], o.line)
 
     comprehension_helper(builder, loop_params, gen_inner_stmts, o.line)
     return set_ops
 
 
 def transform_dictionary_comprehension(builder: IRBuilder, o: DictionaryComprehension) -> Value:
-    d = builder.primitive_op(new_dict_op, [], o.line)
+    d = builder.call_c(dict_new_op, [], o.line)
     loop_params = list(zip(o.indices, o.sequences, o.condlists))
 
     def gen_inner_stmts() -> None:
         k = builder.accept(o.key)
         v = builder.accept(o.value)
-        builder.primitive_op(dict_set_item_op, [d, k, v], o.line)
+        builder.call_c(dict_set_item_op, [d, k, v], o.line)
 
     comprehension_helper(builder, loop_params, gen_inner_stmts, o.line)
     return d
@@ -546,11 +690,18 @@ def transform_slice_expr(builder: IRBuilder, expr: SliceExpr) -> Value:
     args = [get_arg(expr.begin_index),
             get_arg(expr.end_index),
             get_arg(expr.stride)]
-    return builder.primitive_op(new_slice_op, args, expr.line)
+    return builder.call_c(new_slice_op, args, expr.line)
 
 
 def transform_generator_expr(builder: IRBuilder, o: GeneratorExpr) -> Value:
     builder.warning('Treating generator comprehension as list', o.line)
-    return builder.primitive_op(
+    return builder.call_c(
         iter_op, [translate_list_comprehension(builder, o)], o.line
     )
+
+
+def transform_assignment_expr(builder: IRBuilder, o: AssignmentExpr) -> Value:
+    value = builder.accept(o.value)
+    target = builder.get_assignment_target(o.target)
+    builder.assign(target, value, o.line)
+    return value
