@@ -12,22 +12,22 @@ from mypy.nodes import (
     Lvalue, Expression, TupleExpr, CallExpr, RefExpr, GeneratorExpr, ARG_POS, MemberExpr
 )
 from mypyc.ir.ops import (
-    Value, BasicBlock, LoadInt, Branch, Register, AssignmentTarget, TupleGet,
-    AssignmentTargetTuple, TupleSet, BinaryIntOp
+    Value, BasicBlock, Integer, Branch, Register, TupleGet, TupleSet, IntOp
 )
 from mypyc.ir.rtypes import (
     RType, is_short_int_rprimitive, is_list_rprimitive, is_sequence_rprimitive,
-    RTuple, is_dict_rprimitive, short_int_rprimitive
+    RTuple, is_dict_rprimitive, short_int_rprimitive, int_rprimitive
 )
 from mypyc.primitives.registry import CFunctionDescription
 from mypyc.primitives.dict_ops import (
     dict_next_key_op, dict_next_value_op, dict_next_item_op, dict_check_size_op,
     dict_key_iter_op, dict_value_iter_op, dict_item_iter_op
 )
-from mypyc.primitives.list_ops import new_list_op, list_append_op, list_get_item_unsafe_op
+from mypyc.primitives.list_ops import list_append_op, list_get_item_unsafe_op
 from mypyc.primitives.generic_ops import iter_op, next_op
 from mypyc.primitives.exc_ops import no_err_occurred_op
 from mypyc.irbuild.builder import IRBuilder
+from mypyc.irbuild.targets import AssignmentTarget, AssignmentTargetTuple
 
 GenFunc = Callable[[], None]
 
@@ -87,7 +87,7 @@ def for_loop_helper(builder: IRBuilder, index: Lvalue, expr: Expression,
 
 
 def translate_list_comprehension(builder: IRBuilder, gen: GeneratorExpr) -> Value:
-    list_ops = builder.primitive_op(new_list_op, [], gen.line)
+    list_ops = builder.new_list_op([], gen.line)
     loop_params = list(zip(gen.indices, gen.sequences, gen.condlists))
 
     def gen_inner_stmts() -> None:
@@ -199,7 +199,7 @@ def make_for_loop_generator(builder: IRBuilder,
             # seem worth the hassle of supporting dynamically determining which
             # direction of comparison to do.
             if len(expr.args) == 1:
-                start_reg = builder.add(LoadInt(0))
+                start_reg = Integer(0)  # type: Value
                 end_reg = builder.accept(expr.args[0])
             else:
                 start_reg = builder.accept(expr.args[0])
@@ -392,7 +392,7 @@ def unsafe_index(
     # since we want to use __getitem__ if we don't have an unsafe version,
     # so we just check manually.
     if is_list_rprimitive(target.type):
-        return builder.primitive_op(list_get_item_unsafe_op, [target, index], line)
+        return builder.call_c(list_get_item_unsafe_op, [target, index], line)
     else:
         return builder.gen_method_call(target, '__getitem__', [index], None, line)
 
@@ -411,10 +411,10 @@ class ForSequence(ForGenerator):
         # environment class.
         self.expr_target = builder.maybe_spill(expr_reg)
         if not reverse:
-            index_reg = builder.add(LoadInt(0))
+            index_reg = Integer(0)  # type: Value
         else:
             index_reg = builder.binary_op(self.load_len(self.expr_target),
-                                          builder.add(LoadInt(1)), '-', self.line)
+                                          Integer(1), '-', self.line)
         self.index_target = builder.maybe_spill_assignable(index_reg)
         self.target_type = target_type
 
@@ -428,7 +428,7 @@ class ForSequence(ForGenerator):
             # obviously we still need to check against the length,
             # since it could shrink out from under us.
             comparison = builder.binary_op(builder.read(self.index_target, line),
-                                           builder.add(LoadInt(0)), '>=', line)
+                                           Integer(0), '>=', line)
             second_check = BasicBlock()
             builder.add_bool_branch(comparison, second_check, self.loop_exit)
             builder.activate_block(second_check)
@@ -460,9 +460,9 @@ class ForSequence(ForGenerator):
         builder = self.builder
         line = self.line
         step = 1 if not self.reverse else -1
-        add = builder.binary_int_op(short_int_rprimitive,
-                                    builder.read(self.index_target, line),
-                                    builder.add(LoadInt(step)), BinaryIntOp.ADD, line)
+        add = builder.int_op(short_int_rprimitive,
+                             builder.read(self.index_target, line),
+                             Integer(step), IntOp.ADD, line)
         builder.assign(self.index_target, add, line)
 
 
@@ -495,8 +495,8 @@ class ForDictionaryCommon(ForGenerator):
 
         # We add some variables to environment class, so they can be read across yield.
         self.expr_target = builder.maybe_spill(expr_reg)
-        offset_reg = builder.add(LoadInt(0))
-        self.offset_target = builder.maybe_spill_assignable(offset_reg)
+        offset = Integer(0)
+        self.offset_target = builder.maybe_spill_assignable(offset)
         self.size = builder.maybe_spill(self.load_len(self.expr_target))
 
         # For dict class (not a subclass) this is the dictionary itself.
@@ -517,7 +517,7 @@ class ForDictionaryCommon(ForGenerator):
 
         should_continue = builder.add(TupleGet(self.next_tuple, 0, line))
         builder.add(
-            Branch(should_continue, self.body_block, self.loop_exit, Branch.BOOL_EXPR)
+            Branch(should_continue, self.body_block, self.loop_exit, Branch.BOOL)
         )
 
     def gen_step(self) -> None:
@@ -605,7 +605,11 @@ class ForRange(ForGenerator):
         self.end_reg = end_reg
         self.step = step
         self.end_target = builder.maybe_spill(end_reg)
-        index_reg = builder.alloc_temp(start_reg.type)
+        if is_short_int_rprimitive(start_reg.type) and is_short_int_rprimitive(end_reg.type):
+            index_type = short_int_rprimitive
+        else:
+            index_type = int_rprimitive
+        index_reg = Register(index_type)
         builder.assign(index_reg, start_reg, -1)
         self.index_reg = builder.maybe_spill_assignable(index_reg)
         # Initialize loop index to 0. Assert that the index target is assignable.
@@ -630,13 +634,13 @@ class ForRange(ForGenerator):
         # short ints.
         if (is_short_int_rprimitive(self.start_reg.type)
                 and is_short_int_rprimitive(self.end_reg.type)):
-            new_val = builder.binary_int_op(short_int_rprimitive,
-                            builder.read(self.index_reg, line),
-                            builder.add(LoadInt(self.step)), BinaryIntOp.ADD, line)
+            new_val = builder.int_op(short_int_rprimitive,
+                                     builder.read(self.index_reg, line),
+                                     Integer(self.step), IntOp.ADD, line)
 
         else:
             new_val = builder.binary_op(
-                builder.read(self.index_reg, line), builder.add(LoadInt(self.step)), '+', line)
+                builder.read(self.index_reg, line), Integer(self.step), '+', line)
         builder.assign(self.index_reg, new_val, line)
         builder.assign(self.index_target, new_val, line)
 
@@ -648,7 +652,7 @@ class ForInfiniteCounter(ForGenerator):
         builder = self.builder
         # Create a register to store the state of the loop index and
         # initialize this register along with the loop index to 0.
-        zero = builder.add(LoadInt(0))
+        zero = Integer(0)
         self.index_reg = builder.maybe_spill_assignable(zero)
         self.index_target = builder.get_assignment_target(
             self.index)  # type: Union[Register, AssignmentTarget]
@@ -660,9 +664,9 @@ class ForInfiniteCounter(ForGenerator):
         # We can safely assume that the integer is short, since we are not going to wrap
         # around a 63-bit integer.
         # NOTE: This would be questionable if short ints could be 32 bits.
-        new_val = builder.binary_int_op(short_int_rprimitive,
-                builder.read(self.index_reg, line),
-                builder.add(LoadInt(1)), BinaryIntOp.ADD, line)
+        new_val = builder.int_op(short_int_rprimitive,
+                                 builder.read(self.index_reg, line),
+                                 Integer(1), IntOp.ADD, line)
         builder.assign(self.index_reg, new_val, line)
         builder.assign(self.index_target, new_val, line)
 
