@@ -12,13 +12,13 @@ import importlib.util
 from mypy.nodes import (
     Block, ExpressionStmt, ReturnStmt, AssignmentStmt, OperatorAssignmentStmt, IfStmt, WhileStmt,
     ForStmt, BreakStmt, ContinueStmt, RaiseStmt, TryStmt, WithStmt, AssertStmt, DelStmt,
-    Expression, StrExpr, TempNode, Lvalue, Import, ImportFrom, ImportAll
+    Expression, StrExpr, TempNode, Lvalue, Import, ImportFrom, ImportAll, TupleExpr, ListExpr,
+    StarExpr
 )
 
 from mypyc.ir.ops import (
-    Assign, Unreachable, AssignmentTarget, AssignmentTargetRegister, AssignmentTargetIndex,
-    AssignmentTargetAttr, AssignmentTargetTuple, RaiseStandardError, LoadErrorValue,
-    BasicBlock, TupleGet, Value, Register, Branch, NO_TRACEBACK_LINE_NO
+    Assign, Unreachable, RaiseStandardError, LoadErrorValue, BasicBlock, TupleGet, Value, Register,
+    Branch, NO_TRACEBACK_LINE_NO
 )
 from mypyc.ir.rtypes import exc_rtuple
 from mypyc.primitives.generic_ops import py_delattr_op
@@ -28,6 +28,10 @@ from mypyc.primitives.exc_ops import (
     raise_exception_op, reraise_exception_op, error_catch_op, exc_matches_op, restore_exc_info_op,
     get_exc_value_op, keep_propagating_op, get_exc_info_op
 )
+from mypyc.irbuild.targets import (
+    AssignmentTarget, AssignmentTargetRegister, AssignmentTargetIndex, AssignmentTargetAttr,
+    AssignmentTargetTuple
+)
 from mypyc.irbuild.nonlocalcontrol import (
     ExceptNonlocalControl, FinallyNonlocalControl, TryFinallyNonlocalControl
 )
@@ -69,25 +73,47 @@ def transform_return_stmt(builder: IRBuilder, stmt: ReturnStmt) -> None:
 
 
 def transform_assignment_stmt(builder: IRBuilder, stmt: AssignmentStmt) -> None:
-    assert len(stmt.lvalues) >= 1
-    builder.disallow_class_assignments(stmt.lvalues, stmt.line)
-    lvalue = stmt.lvalues[0]
+    lvalues = stmt.lvalues
+    assert len(lvalues) >= 1
+    builder.disallow_class_assignments(lvalues, stmt.line)
+    first_lvalue = lvalues[0]
     if stmt.type and isinstance(stmt.rvalue, TempNode):
         # This is actually a variable annotation without initializer. Don't generate
         # an assignment but we need to call get_assignment_target since it adds a
         # name binding as a side effect.
-        builder.get_assignment_target(lvalue, stmt.line)
+        builder.get_assignment_target(first_lvalue, stmt.line)
+        return
+
+    # Special case multiple assignments like 'x, y = e1, e2'.
+    if (isinstance(first_lvalue, (TupleExpr, ListExpr))
+            and isinstance(stmt.rvalue, (TupleExpr, ListExpr))
+            and len(first_lvalue.items) == len(stmt.rvalue.items)
+            and all(is_simple_lvalue(item) for item in first_lvalue.items)
+            and len(lvalues) == 1):
+        temps = []
+        for right in stmt.rvalue.items:
+            rvalue_reg = builder.accept(right)
+            temp = Register(rvalue_reg.type)
+            builder.assign(temp, rvalue_reg, stmt.line)
+            temps.append(temp)
+        for (left, temp) in zip(first_lvalue.items, temps):
+            assignment_target = builder.get_assignment_target(left)
+            builder.assign(assignment_target, temp, stmt.line)
         return
 
     line = stmt.rvalue.line
     rvalue_reg = builder.accept(stmt.rvalue)
     if builder.non_function_scope() and stmt.is_final_def:
-        builder.init_final_static(lvalue, rvalue_reg)
-    for lvalue in stmt.lvalues:
+        builder.init_final_static(first_lvalue, rvalue_reg)
+    for lvalue in lvalues:
         target = builder.get_assignment_target(lvalue)
         builder.assign(target, rvalue_reg, line)
 
 
+def is_simple_lvalue(expr: Expression) -> bool:
+    return not isinstance(expr, (StarExpr, ListExpr, TupleExpr))
+
+
 def transform_operator_assignment_stmt(builder: IRBuilder, stmt: OperatorAssignmentStmt) -> None:
     """Operator assignment statement such as x += 1"""
     builder.disallow_class_assignments([stmt.lvalue], stmt.line)
@@ -117,6 +143,8 @@ def transform_import(builder: IRBuilder, node: Import) -> None:
         # that mypy couldn't find, since it doesn't analyze module references
         # from those properly.
 
+        # TODO: Don't add local imports to the global namespace
+
         # Miscompiling imports inside of functions, like below in import from.
         if as_name:
             name = as_name
@@ -125,9 +153,11 @@ def transform_import(builder: IRBuilder, node: Import) -> None:
             base = name = node_id.split('.')[0]
 
         # Python 3.7 has a nice 'PyImport_GetModule' function that we can't use :(
-        mod_dict = builder.primitive_op(get_module_dict_op, [], node.line)
+        mod_dict = builder.call_c(get_module_dict_op, [], node.line)
+        # Get top-level module/package object.
         obj = builder.call_c(dict_get_item_op,
                              [mod_dict, builder.load_static_unicode(base)], node.line)
+
         builder.gen_method_call(
             globals, '__setitem__', [builder.load_static_unicode(name), obj],
             result_type=None, line=node.line)
@@ -291,7 +321,7 @@ def transform_try_except(builder: IRBuilder,
             matches = builder.call_c(
                 exc_matches_op, [builder.accept(type)], type.line
             )
-            builder.add(Branch(matches, body_block, next_block, Branch.BOOL_EXPR))
+            builder.add(Branch(matches, body_block, next_block, Branch.BOOL))
             builder.activate_block(body_block)
         if var:
             target = builder.get_assignment_target(var)
@@ -325,7 +355,7 @@ def transform_try_except(builder: IRBuilder,
     # the exception.
     builder.activate_block(double_except_block)
     builder.call_c(restore_exc_info_op, [builder.read(old_exc)], line)
-    builder.primitive_op(keep_propagating_op, [], NO_TRACEBACK_LINE_NO)
+    builder.call_c(keep_propagating_op, [], NO_TRACEBACK_LINE_NO)
     builder.add(Unreachable())
 
     # If present, compile the else body in the obvious way
@@ -376,7 +406,7 @@ def try_finally_entry_blocks(builder: IRBuilder,
                              main_entry: BasicBlock,
                              finally_block: BasicBlock,
                              ret_reg: Optional[Register]) -> Value:
-    old_exc = builder.alloc_temp(exc_rtuple)
+    old_exc = Register(exc_rtuple)
 
     # Entry block for non-exceptional flow
     builder.activate_block(main_entry)
@@ -463,7 +493,7 @@ def try_finally_resolve_control(builder: IRBuilder,
     # If there was an exception, restore again
     builder.activate_block(cleanup_block)
     finally_control.gen_cleanup(builder, -1)
-    builder.primitive_op(keep_propagating_op, [], NO_TRACEBACK_LINE_NO)
+    builder.call_c(keep_propagating_op, [], NO_TRACEBACK_LINE_NO)
     builder.add(Unreachable())
 
     return out_block
@@ -564,7 +594,7 @@ def transform_with(builder: IRBuilder,
     def finally_body() -> None:
         out_block, exit_block = BasicBlock(), BasicBlock()
         builder.add(
-            Branch(builder.read(exc), exit_block, out_block, Branch.BOOL_EXPR)
+            Branch(builder.read(exc), exit_block, out_block, Branch.BOOL)
         )
         builder.activate_block(exit_block)
         none = builder.none_object()
